#include <ntifs.h>

#define IOCTL_GET_PROCESSES_INFO CTL_CODE(FILE_DEVICE_UNKNOWN,4000,METHOD_OUT_DIRECT,GENERIC_READ | GENERIC_WRITE)

typedef NTSTATUS (__stdcall *NtQuerySysInfo) (ULONG,PVOID,ULONG,PULONG);

typedef struct _MY_DEVICE_EXTENSION{	
	NtQuerySysInfo FindInfo;

}MY_DEVICE_EXTENSION, *PMY_DEVICE_EXTENSION;

PDEVICE_OBJECT MyDevice;
UNICODE_STRING pRealSymName;
WCHAR pSymName[] = L"\\??\\INFTOOLS";

VOID DriverUnload(IN PDRIVER_OBJECT TheDriver)
{
	IoDeleteSymbolicLink(&pRealSymName);
	IoDeleteDevice(MyDevice);	
}

NTSTATUS DefaultIrpHandler(IN PDEVICE_OBJECT MyDevice,IN PIRP MyIrp)
{
	MyIrp -> IoStatus.Status = STATUS_SUCCESS;
	MyIrp -> IoStatus.Information = 0;
	IoCompleteRequest(MyIrp,IO_NO_INCREMENT);
	return STATUS_SUCCESS;
}

NTSTATUS DeviceControlHandler(IN PDEVICE_OBJECT MyDevice,IN PIRP MyIrp)
{
	PIO_STACK_LOCATION Stack;
	NTSTATUS Status;
	ULONG OutLength;
	PVOID DataBuffer;
	PMY_DEVICE_EXTENSION pdx;
	NtQuerySysInfo MyCall;

	Stack = IoGetCurrentIrpStackLocation(MyIrp);
	if (Stack -> Parameters.DeviceIoControl.IoControlCode == IOCTL_GET_PROCESSES_INFO)
	{
		OutLength = Stack -> Parameters.DeviceIoControl.OutputBufferLength;
		DataBuffer = MmGetSystemAddressForMdlSafe(MyIrp -> MdlAddress,NormalPagePriority);
		if (DataBuffer == NULL)
		{
			Status = STATUS_INVALID_DEVICE_REQUEST;
			MyIrp -> IoStatus.Status = STATUS_INVALID_DEVICE_REQUEST;
			MyIrp -> IoStatus.Information = 0;
		}
		else
		{
			pdx = (PMY_DEVICE_EXTENSION) (MyDevice -> DeviceExtension);
			MyCall = pdx -> FindInfo;
			if (MyCall == NULL)
			{
				Status = STATUS_INVALID_DEVICE_REQUEST;
				MyIrp -> IoStatus.Status = STATUS_INVALID_DEVICE_REQUEST;
				MyIrp -> IoStatus.Information = 0;
			}
			else
			{
				Status = MyCall(5,DataBuffer,OutLength,NULL);
				MyIrp -> IoStatus.Status = Status;
				if (NT_SUCCESS(Status))
				{
					MyIrp -> IoStatus.Information = OutLength;
				}
				else
				{
					MyIrp -> IoStatus.Information = 0;
				}
			}
		}
	}
	else
	{
		Status = STATUS_INVALID_DEVICE_REQUEST;
		MyIrp -> IoStatus.Status = STATUS_INVALID_DEVICE_REQUEST;
		MyIrp -> IoStatus.Information = 0;
	}
	IoCompleteRequest(MyIrp,IO_NO_INCREMENT);
	return Status;
}

NTSTATUS DriverEntry(IN PDRIVER_OBJECT MyDriver,IN PUNICODE_STRING RegistryStr)
{
	NTSTATUS Status;
	WCHAR pDevName[] = L"\\Device\\INFTOOL";	
	WCHAR ServiceName[] = L"NtQuerySystemInformation";
	UNICODE_STRING pRealDevName;
	UNICODE_STRING FuncName;
	NTSTATUS I;
	PMY_DEVICE_EXTENSION pdx;
	
	Status = STATUS_SUCCESS;
	MyDriver -> DriverUnload = DriverUnload;
	RtlInitUnicodeString(&pRealDevName,pDevName);
	RtlInitUnicodeString(&pRealSymName,pSymName);
	Status = IoCreateDevice(MyDriver,sizeof(MY_DEVICE_EXTENSION),&pRealDevName,FILE_DEVICE_UNKNOWN,0,TRUE,&MyDevice);
	if (NT_SUCCESS(Status))
	{
		pdx = (PMY_DEVICE_EXTENSION) MyDevice -> DeviceExtension;		
		RtlInitUnicodeString(&FuncName,ServiceName);
		pdx -> FindInfo = MmGetSystemRoutineAddress(&FuncName);
		MyDevice -> Flags = MyDevice -> Flags | DO_DIRECT_IO | DO_BUFFERED_IO;
		Status = IoCreateSymbolicLink(&pRealSymName,&pRealDevName);
		if (NT_SUCCESS(Status))
		{
			for (I = 0;I < IRP_MJ_MAXIMUM_FUNCTION;I++)
			{
				MyDriver -> MajorFunction[I] = DefaultIrpHandler;
			}
			MyDriver -> MajorFunction[IRP_MJ_DEVICE_CONTROL] = DeviceControlHandler;
		}
		else
		{
			IoDeleteDevice(MyDevice);
		}
	}
	return Status;
}